﻿/*
** NOTE:
** It doesn't make sense to test the GameModeSeonds field before fixing, like
** the older fixers do. The reason for this is that the in-game bug depends on
** the framerate the game runs at. For example, a GameModeSeonds of >= 131072.0f
** is enough to cause the bug if you run the game at 240Hz, while it has to be
** at least 524288.0f if you run it at 60Hz.
** oblanimfix resets this field to 0.0f regardless, which fixes the bug for all
** players.
*/

#define _CRT_SECURE_NO_WARNINGS /* prevents some Visual Studio warnings */

#include <stdint.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifdef _WIN32
#include <windows.h>
typedef wchar_t UNICHAR;
#define UNICHAR_FOPEN(a, b) _wfopen(a, L ## b)
#define UNICHAR_STRCPY(a, b) wcscpy(a, b)
#define UNICHAR_STRCAT(a, b) wcscat(a, L ## b)
#define UNICHAR_STRLEN(a) wcslen(a)
#else
typedef char UNICHAR;
#define UNICHAR_FOPEN(a, b) fopen(a, b)
#define UNICHAR_STRCPY(a, b) strcpy(a, b)
#define UNICHAR_STRCAT(a, b) strcat(a, b)
#define UNICHAR_STRLEN(a) strlen(a)
#defi
#endif

#define SWAP16(x) \
( \
	(((uint16_t)((x) & 0x00FF)) << 8) | \
	(((uint16_t)((x) & 0xFF00)) >> 8)   \
)

#define SWAP32(x) \
( \
	(((uint32_t)((x) & 0x000000FF)) << 24) | \
	(((uint32_t)((x) & 0x0000FF00)) <<  8) | \
	(((uint32_t)((x) & 0x00FF0000)) >>  8) | \
	(((uint32_t)((x) & 0xFF000000)) >> 24)   \
)

static bool bigEndian;
static bool read8(FILE *f, uint8_t *out);
static bool readLE16(FILE *f, uint16_t *out);
static bool readLE32(FILE *f, uint32_t *out);
static bool createBak(UNICHAR *filename, FILE *f);

int main(int argc, char* argv[])
{
	FILE *f = NULL;
#ifdef WIN32
	UNICHAR **argvw = NULL;
	bool didAllocateArgvW = false;
#endif
	// detect endianness
	volatile uint32_t endiantest32 = 0xFF;
	bigEndian = (*(uint8_t *)&endiantest32 != 0xFF);

	printf("Oblivion Animation Bug Fixer (oblanimfix) by 8bitbubsy, 2023\n");
	printf("============================================================\n\n");

	if (argc != 2)
	{
		printf("Usage: oblanimfix <savegame.ess>\n");
#ifdef _WIN32
		printf("       (or drag and drop a .ess file onto this executable)\n");
		printf("\n");
		printf("This program can only handle save files from the PC version.\n");
		printf("\n");
		printf("Tip: The save files for Oblivion are located at\n");
		printf("     Documents\\My Games\\Oblivion\\Saves\\\n");
		printf("\n");
		system("PAUSE");
#endif
		return -1;
	}
	
	UNICHAR *filename;
#ifdef _WIN32
	/* Read input filename in widechar type (Windows is the ultimate portability killer!)
	** (yes, I know I can just do int wmain(...), but MinGW-W32 doesn't like this)
	*/
	int32_t numArgs;
	argvw = CommandLineToArgvW(GetCommandLineW(), &numArgs);
	if (argvw == NULL || numArgs != 2)
	{
		printf("ERROR: Couldn't read command line argument!\n");
		goto Error;
	}
	didAllocateArgvW = true;
	filename = argvw[1];
#else
	filename = argv[1];
#endif

	f = UNICHAR_FOPEN(filename, "r+b");
	if (f == NULL)
	{
		printf("ERROR: Couldn't open input file for reading!\n");
		goto Error;
	}

	char ID[12];
	uint8_t majVersion, minVersion;

	if (fread(ID, 12, 1, f) != 1) goto IOError;
	if (!read8(f, &majVersion)) goto IOError;
	if (!read8(f, &minVersion)) goto IOError;

	if (memcmp(ID, "TES4SAVEGAME", 12) != 0 || majVersion != 0)
	{
		printf("ERROR: This is not an Oblivion save game!\n");
		goto Error;
	}

	if (minVersion != 125 && minVersion != 126)
	{
		printf("ERROR: This Oblivion save game version is not compatible!\n");
		goto Error;
	}

	printf("Creating backup of save... ");
	if (!createBak(filename, f))
	{
		printf("FAILED! Exiting...\n");
		goto Error;
	}
	printf("Ok.\n");

	uint8_t data8;
	uint16_t data16;
	uint32_t data32;

	// seek to save game header
	fseek(f, 42, SEEK_SET);
	if (!read8(f, &data8)) goto IOError;
	fseek(f, data8 + 2, SEEK_CUR);
	if (!read8(f, &data8)) goto IOError;
	fseek(f, data8, SEEK_CUR);
	fseek(f, 24, SEEK_CUR);
	if (!readLE32(f, &data32)) goto IOError;
	fseek(f, data32, SEEK_CUR);

	// skip plugins
	uint8_t numPlugins;
	if (!read8(f, &numPlugins)) goto IOError;
	for (uint32_t i = 0; i < numPlugins; i++)
	{
		if (!read8(f, &data8)) goto IOError;
		fseek(f, data8, SEEK_CUR);
	}

	// we're now in the global header, skip to gameModeSeconds
	fseek(f, 40, SEEK_CUR);
	if (!readLE16(f, &data16)) goto IOError;
	fseek(f, data16*8, SEEK_CUR);
	if (!readLE16(f, &data16)) goto IOError;
	fseek(f, data16-4, SEEK_CUR);

	// reset gameModeSeconds field (IEEE-754 single-precision float)
	const uint32_t zero32 = 0;
	if (fwrite(&zero32, 4, 1, f) != 1)
	{
		printf("ERROR: Couldn't modify the GameModeSeconds field!\n");
		printf("The save may be damaged now... Please locate the backup (.bak) file.\n");
		goto Error;
	}

	fclose(f);
	printf("The GameModeSeconds field has been successfully reset!\n");

#ifdef _WIN32
	free(*argvw);
	printf("\n");
	system("PAUSE");
#endif
	return 0;

IOError:
	printf("ERROR: I/O error while reading from file!\n");
Error:
	if (f != NULL) fclose(f);

#ifdef _WIN32
	if (didAllocateArgvW) free(*argvw);
	printf("\n");
	system("PAUSE");
#endif

	return 1;
}

static bool read8(FILE *f, uint8_t *out)
{
	uint8_t data8;
	if (fread(&data8, 1, 1, f) != 1)
		return false;

	*out = data8;
	return true;
}

static bool readLE16(FILE *f, uint16_t *out)
{
	uint16_t data16;
	if (fread(&data16, 2, 1, f) != 1)
		return false;

	if (bigEndian)
		data16 = SWAP16(data16);

	*out = data16;
	return true;
}

static bool readLE32(FILE *f, uint32_t *out)
{
	uint32_t data32;
	if (fread(&data32, 4, 1, f) != 1)
		return false;

	if (bigEndian)
		data32 = SWAP32(data32);

	*out = data32;
	return true;
}

static bool createBak(UNICHAR *filename, FILE *f)
{
	uint8_t buffer[32768];
	FILE *fBak = NULL;
	UNICHAR *bakFilename = NULL;

	if (filename == NULL || f == NULL)
		return false;

	bakFilename = (UNICHAR *)malloc((UNICHAR_STRLEN(filename) + 1 + 3 + 1) * sizeof (UNICHAR));
	if (bakFilename == NULL)
		goto Error;

	UNICHAR_STRCPY(bakFilename, filename);
	UNICHAR_STRCAT(bakFilename, ".bak");

	fBak = UNICHAR_FOPEN(bakFilename, "wb");
	if (fBak == NULL)
		goto Error;

	const uint32_t oldPos = (uint32_t)ftell(f);

	fseek(f, 0, SEEK_END);
	const uint32_t filesize = (uint32_t)ftell(f);
	fseek(f, 0, SEEK_SET);
	
	uint32_t bytesLeft = filesize;
	while (bytesLeft > 0)
	{
		uint32_t bytesTodo = sizeof (buffer);
		if (bytesTodo > bytesLeft)
			bytesTodo = bytesLeft;
			
		if ( fread(buffer, 1, bytesTodo, f)    != bytesTodo) goto Error;
		if (fwrite(buffer, 1, bytesTodo, fBak) != bytesTodo) goto Error;
		
		bytesLeft -= bytesTodo;
	}
	
	fclose(fBak);
	free(bakFilename);

	fseek(f, oldPos, SEEK_SET);
	return true;
	
Error:
	if (fBak != NULL) fclose(fBak);
	if (bakFilename != NULL) free(bakFilename);
	
	fseek(f, oldPos, SEEK_SET);
	return false;
}
